/*
 * Yacc syntax for logic description language
 * P. Mackerras  10-Jun-89
 */

/* declarations */

%{
#include <stdlib.h>
#include "logsyn.h"
%}

%union {
	int		y_int;
	char		*y_name;
	struct ident	*y_ident;
	struct attr	*y_attr;
	struct signal	*y_signal;
	struct expr	*y_expr;
}

%term INPUT, OUTPUT, INTERNAL, KFAST, ACTIVE_LOW, ACTIVE_HIGH
%term CLOCKED_BY, LATCHED_BY, ENABLED_BY, RESET_BY, HAZARD_FREE
%term KGROUP, DEFINE, KFUNCTION, DEFINE_PAL
%term TRUTH_TABLE, SUM_OF_PRODUCTS, STATE_MACHINE, ASYNCHRONOUS
%term TRUE, FALSE, ALL_STATES, OTHER_STATES, PIN_NUMBER, INCLUDE
%term IF, THEN, ELSIF, ELSE, ENDIF
%term BECOMES, NOTEQUAL, EQUAL, IMPLIES, LESSEQ, GTREQ, LESS, GTR
%term <y_name> IDENT, STRING
%term <y_int> HEXNUM

/* operator precedences - lowest to highest */
%left '^'		/* EXCLUSIVE OR operator */
%left '|'		/* OR operator */
%left '&', '#'		/* AND and AND NOT operators */
%nonassoc EQUAL, NOTEQUAL, LESS, LESSEQ, GTR, GTREQ
%nonassoc '~'		/* NOT operator */

%start logic_program

/* types of non-terminals */
%type <y_attr>		signal_decl, attributes, attribute
%type <y_ident>		new_signal
%type <y_int>		number, digit, bit, opt_async, assigned
%type <y_expr>		variable, signal_group, signal_list, signal_name,
			expr, elses, expr_group, expr_list, ident_expr,
			truth_table, sum_of_products, function_use,
			bit_group, bit_expr, transitions, trans_elses,
			bit_sequence, expr_list_elt, ix_expr, opt_signal_list
%type <y_int>		pin_attribute, pin_decl, opt_fast
%type <y_expr>		opt_pin_list, pin_list, pin_list_elt

%%
/* rules */

logic_program	:
		| logic_program declaration
		;

declaration	: ';'
		| signal_decl ';'	{ }
		| group_decl ';'
		| function_decl ';'
		| state_machine ';'
		| equation ';'
		| inclusion
		| pal_spec ';'
		;

signal_decl	: attributes new_signal
					{ signal_decl($1, $2); $$ = $1; }
		| signal_decl ',' new_signal
					{ signal_decl($1, $3); $$ = $1; }
		;

attributes	: attribute
		| attributes attribute
					{ $$ = add_attrs($1, $2); }
		;

attribute	: ACTIVE_HIGH		{ $$ = new_attr(HIGH_TRUE, NULL); }
		| ACTIVE_LOW		{ $$ = new_attr(LOW_TRUE, NULL); }
		| INPUT			{ $$ = new_attr(INPUT_BIT, NULL); }
		| OUTPUT		{ $$ = new_attr(OUTPUT_BIT, NULL); }
		| HAZARD_FREE opt_signal_list
					{ $$ = new_attr(NO_HAZARD, $2); }
		| CLOCKED_BY '(' expr ')'
					{ $$ = new_attr(CLOCKED, $3); }
		| LATCHED_BY '(' expr ')'
 					{ $$ = new_attr(LATCHED, $3); }
		| ENABLED_BY '(' expr ')'
					{ $$ = new_attr(TRISTATE, $3); }
		| KFAST			{ $$ = new_attr(CRIT_PATH, NULL); }
		| RESET_BY '(' expr ')'
					{ $$ = new_attr(ASYNC_RESET, $3); }
		;

new_signal	: IDENT
					{ $$ = new_signal($1, 0, 0); }
		| IDENT '[' number ':' number ']'
					{ $$ = new_signal($1, $3, $5); }
		| IDENT '<' number ':' number '>'
					{ $$ = new_signal($1, $3, $5); }
		;

number		: digit
		| number digit		{ $$ = $1 * 10 + $2; }
		;

digit		: '0'			{ $$ = 0; }
		| '1'			{ $$ = 1; }
		| '2'			{ $$ = 2; }
		| '3'			{ $$ = 3; }
		| '4'			{ $$ = 4; }
		| '5'			{ $$ = 5; }
		| '6'			{ $$ = 6; }
		| '7'			{ $$ = 7; }
		| '8'			{ $$ = 8; }
		| '9'			{ $$ = 9; }
		;

equation	: variable assigned expr
					{ define_vble($1, $3); }
		;

variable	: signal_name
		| signal_group
		;

signal_group	: '<' signal_list '>'	{ $$ = $2; }
		| '<' '>'		{ $$ = NULL; }
		;

signal_list	: signal_name
		| signal_list signal_name
					{ $$ = expr(',', $1, $2); }
		| signal_list ',' signal_name
					{ $$ = expr(',', $1, $3); }
		;

opt_signal_list	: /* empty */		{ $$ = NULL; }
		| '(' signal_list ')'	{ $$ = $2; }
		;

signal_name	: IDENT			{ $$ = signal_expr($1, -1, -1); }
		| IDENT '[' number ']'	{ $$ = signal_expr($1, $3, $3); }
		| IDENT '<' number '>'	{ $$ = signal_expr($1, $3, $3); }
		| IDENT '[' number ':' number ']'
					{ $$ = signal_expr($1, $3, $5); }
		| IDENT '<' number ':' number '>'
					{ $$ = signal_expr($1, $3, $5); }
		;

ident_expr	: IDENT			{ $$ = ident_expr($1, -1, -1); }
		| IDENT '[' number ']'	{ $$ = ident_expr($1, $3, $3); }
		| IDENT '<' number '>'	{ $$ = ident_expr($1, $3, $3); }
		| IDENT '[' number ':' number ']'
					{ $$ = ident_expr($1, $3, $5); }
		| IDENT '<' number ':' number '>'
					{ $$ = ident_expr($1, $3, $5); }
		;

assigned	: '='			{ $$ = 0; }
		| BECOMES		{ $$ = 1; }
		;

group_decl	: KGROUP IDENT assigned signal_group
					{ new_group($2, $4); }
		;

expr		: bit			{ $$ = bit_expr($1); }
		| ident_expr
		| HEXNUM ':' number	{ $$ = hexnum_expr($1, $3); }
		| '~' expr		{ $$ = expr('~', $2, NULL); }
		| expr '&' expr		{ $$ = expr('&', $1, $3); }
		| expr '#' expr		{ $$ = expr('#', $1, $3); }
		| expr '|' expr		{ $$ = expr('|', $1, $3); }
		| expr '^' expr		{ $$ = expr('^', $1, $3); }
		| expr EQUAL expr	{ $$ = expr(EQUAL, $1, $3); }
		| expr NOTEQUAL expr	{ $$ = expr(NOTEQUAL, $1, $3); }
		| expr LESS expr	{ $$ = expr(LESS, $1, $3); }
		| expr GTR expr		{ $$ = expr(GTR, $1, $3); }
		| expr LESSEQ expr	{ $$ = expr(LESSEQ, $1, $3); }
		| expr GTREQ expr	{ $$ = expr(GTREQ, $1, $3); }
		| ix_expr
		| ix_expr '[' number ']'
					{ $$ = expr_select($1, $3, $3); }
		| ix_expr '<' number '>'
					{ $$ = expr_select($1, $3, $3); }
		| ix_expr '[' number ':' number ']'
					{ $$ = expr_select($1, $3, $5); }
		| ix_expr '<' number ':' number '>'
					{ $$ = expr_select($1, $3, $5); }
		;

ix_expr		: function_use
		| truth_table
		| sum_of_products
		| '(' expr ')'		{ $$ = $2; }
		| expr_group
		| IF expr THEN expr elses
				{ $$ = expr(IF, $2, expr(ELSE, $4, $5)); }

elses		: ENDIF			{ $$ = NULL; }
		| ELSE expr ENDIF	{ $$ = $2; }
		| ELSIF expr THEN expr elses
				{ $$ = expr(IF, $2, expr(ELSE, $4, $5)); }
		;

bit		: '0'			{ $$ = Z; }
		| '1'			{ $$ = O; }
		| FALSE			{ $$ = Z; }
		| TRUE			{ $$ = O; }
		| '%'			{ $$ = X; }
		| '?'			{ $$ = N; }
		;

bit_sequence	: bit bit
				{ $$ = expr(',', bit_expr($1), bit_expr($2)); }
		| bit_sequence bit	{ $$ = expr(',', $1, bit_expr($2)); }
		;

expr_group	: '<' expr_list '>'	{ $$ = $2; }
		;

expr_list	: expr_list_elt
		| expr_list ',' expr_list_elt
					{ $$ = expr(',', $1, $3); }
		;

expr_list_elt	: expr
		| bit_sequence
		;

truth_table	: TRUTH_TABLE '[' number ']' expr_group
					{ new_tt($5, $3); }
		  '{' group_pair_list '}'
					{ $$ = tt_expr(); }
		;

sum_of_products	: SUM_OF_PRODUCTS '[' number ']' expr_group
					{ new_tt($5, $3); }
		  '{' group_pair_list '}'
					{ $$ = sp_expr(); }
		;

group_pair_list	:
		| group_pair_list io_bits
					{ add_tt(); }
		;

io_bits		: 			{ new_bitlist(); }
		  '<' bit_list '>'	{ check_ins(); }
		  '<' bit_list '>'	{ check_outs(); }
		;

bit_list	:
		| bit_list bit		{ add_bit($2); }
		;

function_use	: IDENT '(' expr_list ')'
					{ $$ = use_func($1, $3); }
		;

function_decl	: KFUNCTION IDENT	{ new_level(); }
		  opt_formal_par assigned expr
					{ new_func($2, $6); }
		;

opt_formal_par	:
		| '(' formal_params ')'
		;

formal_params	: new_signal		{ new_param($1); }
		| formal_params ',' new_signal
					{ new_param($3); }
		;

state_machine	: opt_async STATE_MACHINE IMPLIES variable
					{ state_outputs($4); }
		  '{' state_list '}'
					{ state_machine($1); }
		;

opt_async	:			{ $$ = 0; }
		| ASYNCHRONOUS		{ $$ = 1; }
		;

state_list	:
		| state_list state_spec ';'
		;

state_spec	: ALL_STATES ':' transitions
					{ state_indep($3); }
		| IDENT ':' signal_group transitions
					{ state_sig($1, $3, $4); }
		| IDENT ':' bit_group transitions
					{ state_bit($1, $3, $4); }
		| OTHER_STATES ':' transitions
					{ state_dflt($3); }
		;

bit_group	: '<' bit_expr '>'	{ $$ = $2; }
		;

bit_expr	: bit			{ $$ = bit_expr($1); }
		| bit_expr bit		{ $$ = expr(',', $1, bit_expr($2)); }
		;

transitions	: IMPLIES IDENT		{ $$ = goto_state($2); }
		| IF expr THEN transitions trans_elses
				{ $$ = expr(IF, $2, expr(ELSE, $4, $5)); }
		;

trans_elses	: ENDIF			{ $$ = NULL; }
		| ELSE transitions ENDIF
					{ $$ = $2; }
		| ELSIF expr THEN transitions trans_elses
				{ $$ = expr(IF, $2, expr(ELSE, $4, $5)); }
		;

inclusion	: INCLUDE STRING ';'	{ do_include($2); }

pal_spec	: DEFINE_PAL IDENT '=' STRING
					{ start_pal($2, $4); }
		  '{' pin_assts '}'
					{ do_pal(); }
		;

pin_assts	:
		| pin_assts pin_decl ';'
		;

pin_decl	: pin_attribute signal_name opt_pin_list
					{ pin_decl($1, $2, $3); $$ = $1; }
		| pin_decl ',' signal_name opt_pin_list
					{ pin_decl($1, $3, $4); $$ = $1; }
		;

pin_attribute	: opt_fast INPUT	{ $$ = $1 | INPUT_BIT; }
		| opt_fast OUTPUT	{ $$ = $1 | OUTPUT_BIT; }
		| opt_fast CLOCKED_BY	{ $$ = $1 | CLOCKED; }
		| opt_fast ENABLED_BY	{ $$ = $1 | TRISTATE; }
		| opt_fast RESET_BY	{ $$ = $1 | ASYNC_RESET; }
		| INTERNAL		{ $$ = INTERNAL_SIG; }
		;

opt_fast	:			{ $$ = 0; }
		| KFAST			{ $$ = CRIT_PATH; }
		;

opt_pin_list	:			{ $$ = NULL; }
		| '(' pin_list ')'	{ $$ = $2; }
		;

pin_list	: pin_list_elt
		| pin_list ',' pin_list_elt
					{ $$ = expr(',', $1, $3); }
		;

pin_list_elt	: number		{ $$ = pin_expr($1, $1); }
		| number ':' number	{ $$ = pin_expr($1, $3); }
		;

%%
/* programs */

#include "lex.yy.c"

